#!/usr/bin/perl
use warnings;
use strict;
use Digest::CRC qw(crc64 crc32 crc16 crcccitt crc crc8 crcopenpgparmor crcccitt_hex crcccitt_base64);
use Log::Dispatch;
use Config::IniFiles;
use Net::Ping;
use POE;
use POE::Wheel::ReadWrite;
use POE::Wheel::ReadLine;
use Symbol qw(gensym);
use Device::SerialPort;
use POE::Filter::Line;

## %ConfigFile and %Config container for ini file
my $ConfigFile = "motestation.cfg";
my %Config;
my $log;

LogSetup();
ConfigInit();


	POE::Session->create(
	  inline_states => {
	    _start      => \&setup_device,
	    got_port    => \&display_port_data,
	    got_console => \&transmit_console_data,
	    got_error   => \&handle_errors,
	  },
	);
	POE::Kernel->run();
	exit 0;

sub setup_device {
  my ($kernel, $heap) = @_[KERNEL, HEAP];
  printki(":: poe_socket_setup Section - init\n");


  # Open a serial port, and tie it to a file handle for POE.
  my $handle = gensym();
  my $port = tie(*$handle, "Device::SerialPort", $Config{SerialPort});
  die "can't open port: $!" unless $port;
  $port->datatype('raw');
  $port->baudrate($Config{SerialSpeed});
  $port->databits($Config{SerialDataBits});
  $port->parity($Config{SerialParity});
  $port->stopbits($Config{SerialStopBits});
  $port->handshake("rts");
  $port->write_settings();

  # Start interacting with the GPS.
  $heap->{port}       = $port;
  $heap->{port_wheel} = POE::Wheel::ReadWrite->new(
    Handle => $handle,
    Filter => POE::Filter::Line->new(
      InputLiteral  => "\x0D\x0A",    # Received line endings.
      OutputLiteral => "\x0D",        # Sent line endings.
    ),
    InputEvent => "got_port",
    ErrorEvent => "got_error",
  );

  # Start a wheel to interact with the console, and prompt the user.
  $heap->{console} = POE::Wheel::ReadLine->new(InputEvent => "got_console",);
  $heap->{console}->put("Press ^D to stop.");
  $heap->{console}->get("Ready: ");
}

# Port data (lines, separated by CRLF) are displayed on the console.
sub display_port_data {
  my ($heap, $data) = @_[HEAP, ARG0];
  my $string = $data;

  $string =~ s/\x7D\x5E/"\x7E"/eg;
  $string =~ s/\x7D\x5D/"\x7D"/eg;
  #$string =~ s/(.)/sprintf(" %02X",ord($1))/eg;
  #$string =~ s/\n/" 0A"/eg;

  if ($string) {
  	my $wocrcstr = substr($data,0,-2);
  	my $crcstring = substr($data,0,-2);
	#$crcstring =~ s/\x7D\x5E/"\x7E"/eg;
	#$crcstring =~ s/\x7D\x5D/"\x7D"/eg;
  	my $crc = crc16($crcstring);
	#$crcstring =~ s/(.)/sprintf(" %02X",ord($1))/eg;
	#$crcstring =~ s/\n/" 0A"/eg;
	#$heap->{console}->put("> O: ".$string."<");
	#$heap->{console}->put("> C: ".$crcstring."<");
	#$heap->{console}->put(">".sprintf(" %04X -    ",$crc).$crcstring."<");
	#$heap->{console}->put("> R: ".sprintf(" %04X - %s\n",unpack('n',pack('S',calcCRC($crcstring))),$crcstring));
	#$heap->{console}->put("> R: ".sprintf(" %04X - %s\n",unpack('n',pack('S',calcCRC($crcstring))),$crcstring));
	$heap->{console}->put("> I: ".hexdump($string)."<");
	$heap->{console}->put("> C: ".hexdump($crcstring)." <".doCRC($crcstring).">");
	#$heap->{console}->put("> T: ".doCRC($crcstring)."<");

  }
}

# Console input is sent to the device.
sub transmit_console_data {
  my ($heap, $input) = @_[HEAP, ARG0];
  if (defined $input) {
    $heap->{console}->addhistory($input);
    $heap->{port_wheel}->put($input);
    $heap->{console}->get("Ready: ");

    # Clearing $! after $serial_port_wheel->put() seems to work around
    # an issue in Device::SerialPort 1.000_002.
    $! = 0;
    return;
  }
  $heap->{console}->put("Bye!");
  delete $heap->{port_wheel};
  delete $heap->{console};
}

# Error on the serial port.  Shut down.
sub handle_errors {
  my $heap = $_[HEAP];
  $heap->{console}->put("$_[ARG0] error $_[ARG1]: $_[ARG2]");
  $heap->{console}->put("bye!");
  delete $heap->{console};
  delete $heap->{port_wheel};
}

sub LogSetup {
	$log = Log::Dispatch->new( outputs => [[ 'Screen', min_level => 'info' ]]);
}

sub ConfigInit {
	my %ini;
	my $key ='';
	my $value='';
	printki(":: ConfigInit Section - init\n");
	tie %ini, 'Config::IniFiles', (-file => $ConfigFile);
	%Config = %{$ini{"STATION"}};
	while (($key, $value) = each %Config){
		printki ("    read ini: ".$key." = ".$value."\n");
	}
	printki(":: ConfigInit Section - done\n");
}


sub printki {
	$log->info(@_);
}

sub printke {
	$log->error(@_);
}

sub diek {
	printki("Fatal error :: ");
die(@_);
}

sub calcCRC {
	my $acum=0;
	foreach my $byte (split //, $_[0]) {
	$acum=calcByteCRC($acum,$byte);
	}
	return $acum;
}

sub calcByteCRC {
	my $crc  = $_[0];
	my $byte = ord ($_[1]);

	$crc = unpack ('S',pack('S', $crc >> 8)) | unpack('S',pack('S',$crc << 8));
	$crc = unpack ('S',pack('S',$crc ^=  $byte));
	$crc = unpack ('S',pack('S',$crc ^= ($crc & 0xff) >> 4));
	$crc = unpack ('S',pack('S',$crc ^=  $crc << 12));
	$crc = unpack ('S',pack('S',$crc ^= ($crc & 0xff) << 5));
	return $crc;
}

sub hexdump {
	my $hexstr = $_[0];
	$hexstr =~ s/(.)/sprintf(" %02X",ord($1))/eg;
	$hexstr =~ s/\n/" 0A"/eg;
	return $hexstr;

}
sub doCRC {
	my $istr = $_[0];

	#printf(" %04X -",unpack('n',pack('S',calcCRC($istr))));
	return sprintf("%04X",unpack('n',pack('S',calcCRC($istr))));

}
